'''
This module contains several self-contained functions for general purposes
'''

def disable_widget(widgets):
    for widget in widgets:
        if widget:
            widget.setEnabled(False)   

def enable_widget(widgets):
    for widget in widgets:
        if widget:
            widget.setEnabled(True)  

def convert_fraction_to_float(fraction):
    #code adapted from https://stackoverflow.com/questions/1806278/convert-fraction-to-float
    try:
        return float(fraction)
    except ValueError:
        num, denom = fraction.split('/')
        try:
            leading, num = num.split(' ')
            whole = float(leading)
        except ValueError:
            whole = 0
        frac = float(num) / float(denom)
        return whole - frac if whole < 0 else whole + frac

def check_list_is_valid_formula(formula,valid_variable_names):
    flag = 1
    msg = ''
    for elem in formula:
        if type(elem) == str:
            if not ((elem in valid_variable_names) or (elem in ['+','-','*','/'])):
                flag = 0
                msg = "One of the elements ('" + elem + "') of the list is neither a valid data name (in the format dev#.DataName) nor a valid mathematical operation ('+', '-', '*', '/')"
                break
        elif isinstance(elem, list):
            msg, flag = CheckListIsValidFormula(elem,valid_variable_names)
        else:
            flag = 0
            msg = 'One of the elements of the list is neither a string nor a list'
            break
    return msg, flag

def evaluate_formula(formula,variables):
    #formula is a structured list
    #variabes is a dictionary of "NameVariable":ValueVariable
    value = 0
    op = '+'
    for elem in formula:
        if elem in ['+','-','*','/']:
            op = elem
        else: 
            if isinstance(elem, list):
                NewValue = evaluate_formula(elem,variables)
            else:
                NewValue =  variables[elem]
            value = operation(value,NewValue,op)
    return value

def operation(Value1,Value2,op):
    if op == '+':
        Value = Value1 + Value2
    if op == '*':
        Value = Value1 * Value2
    if op== '-':
        Value = Value1 - Value2
    if op == '/':
        try:
            Value = Value1 / Value2
        except ZeroDivisionError:
            Value = float('nan') 
    return Value


import os
import PyQt5.QtWidgets as Qt# QApplication, QWidget, QMainWindow, QPushButton, QHBoxLayout
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import logging
import ergastirio.utils
import ergastirio.styles
import copy

graphics_dir = os.path.join(os.path.dirname(__file__), 'graphics')

'''
The panels defined in this module are meant to be GUIs for other classes, i.e. their events are directly connected to methods of other objects and they dont have
any relevant "internal logic"
They do not have their own logger, and they do not produce any logging event. Rather, they are "directly" connected to methods and functionalities
of the experiment object via PyQt5 signals
'''


############################################################################################################
## status_selector                                                                                        # 
############################################################################################################

class status_selector():

    def __init__(self, parent, experiment):
        # parent        = a QWidget (or QMainWindow) object that will be the parent of this gui
        # experiment    = an experiment() object, whose acquisition will be controlled by this panel

        self.parent = parent
        self.experiment  = experiment 

        ### Connect signals from model to event slots of this GUI
        #self.experiment.sig_list_ramps.connect(self.on_list_ramps_updated)
        #self.experiment.sig_single_set_acquisition.connect(self.on_single_set_acquisition_change)
        #self.experiment.sig_triggered_acquisition.connect(self.on_triggered_acquisition_change)
        #self.experiment.sig_trigger_modality.connect(self.on_trigger_modality_change)
        #self.experiment.sig_trigger_instrument.connect(self.on_trigger_instrument_change)
        #self.experiment.sig_time_interval_internal_trigger.connect(self.on_time_inteval_internal_trigger_change)
        #self.experiment.sig_time_interval_multiple_acquisitions_per_set.connect(self.on_time_interval_multiple_acquisitions_per_set_change)
        #self.experiment.sig_numb_acquisitions_per_set.connect(self.on_numb_acquisitions_per_trigger_change)
        #self.experiment.sig_make_average_set_acquisition.connect(self.on_change_make_average_set_acquisition)
        self.experiment.sig_list_instruments.connect(self.on_list_instruments_updated)
        #self.experiment.sig_trigger_instrument.connect(self.on_trigger_instrument_changed)
        #self.experiment.sig_trigger_instrument_delay.connect(self.on_trigger_delay_changed)

        #self.doing_triggered_acquisition = False    # Flag used to keep track a triggered acquisition is being performed. If this is true, some widget will not be re-enabled when self.experiment.SIG_SINGLE_SET_ACQUISITION_ENDED is sent
        #                                            # but only when self.experiment.SIG_TRIGGERED_ACQUISITION_ENDED is sent

    def create_gui(self): 
        self.create_widgets()
        self.connect_widgets_events_to_functions()

    def create_widgets(self):
        """
        - hbox                                          (QHBoxLayout)

        """

        self.vbox = Qt.QVBoxLayout()
        self.label_Info = Qt.QLabel("This panal allows to create experiment 'states', defined as list of properties of one or more instruments and their values." \
                                     "Once a state is defined, it can be quickly re-created by double clicking on it. " \
                                     "This will automatically set the values of all the properties.")
        self.label_Info.setWordWrap(True) 
        self.tableRamps = Qt.QTableWidget()
        self.tableRamps.setColumnCount(3)
        self.tableRamps.setHorizontalHeaderLabels(["Instrument","Property","Value"])
        header = self.tableRamps.horizontalHeader()
        header.setSectionResizeMode(Qt.QHeaderView.ResizeToContents)       
        #header.setSectionResizeMode(0, Qt.QHeaderView.Stretch)
        self.vbox.addWidget(self.label_Info)
        self.vbox.addWidget(self.tableRamps)

        ### Assign the layout hbox to the widget defined in self.parent
        self.parent.setLayout(self.vbox) 
        self.parent.resize(self.parent.minimumSize())

        
        return
    
    def connect_widgets_events_to_functions(self):
        ### Connect Widget Events to Event Functions
        if hasattr(self,'list_combo_ramps'):
            for index,combo_box in enumerate(self.list_combo_ramps):
                self.list_combo_ramps[index].currentIndexChanged.connect(lambda index_newvalue, index=index : self.change_combo_ChildRamp(index_newvalue, index_combobox=index))
        return self

###########################################################################################################
### Event Slots. They are normally triggered by signals from the model, and change the GUI accordingly  ###
###########################################################################################################

    def on_list_instruments_updated(self, list_instruments):
        print(list_instruments)

    def on_list_ramps_updated(self,list_ramps):
        self.list_ramps = list_ramps #store for later use
        self.tableRamps.clearContents()

        # Create a list of ramps for the comboboxes 
        combobox_items = []
        for index,ramp in enumerate(list_ramps):
            combobox_items.append(f"{str(ramp[1])} -> {str(id(ramp[2]))}")
        combobox_items.append('None')

        self.list_combo_ramps = [] #this will contain the combobox for each row of the table
#
        self.tableRamps.setRowCount(len(list_ramps)) 
        for index,ramp in enumerate(list_ramps):
            item0 = Qt.QTableWidgetItem(str(ramp[1]))
            item1 = Qt.QTableWidgetItem(str(id(ramp[2])))
            self.tableRamps.setItem(index,0,item0)
            self.tableRamps.setItem(index,1,item1)
            combo_ramps = Qt.QComboBox()
            combo_ramps.addItems(combobox_items)
            self.tableRamps.setCellWidget(index,2,combo_ramps)
            self.list_combo_ramps.append(combo_ramps)
            
            #If the ramp corresponding to this row has child, we set the combobox to the corresponding child
            if ramp[3] == -1:
                combo_ramps.setCurrentIndex(len(combobox_items)-1) # This will initialize the combobox to 'None'
            else:
                combo_ramps.setCurrentIndex(ramp[3])

        #populating the table with combobox will create new widgets, thus we need to call again connect_widgets_events_to_functions
        self.connect_widgets_events_to_functions()


#######################
### END Event Slots ###
#######################

###################################################################################################################################################
### GUI Events Functions. They are triggered by direct interaction with the GUI, and they call methods of the experiment (i.e. the model) object.###
###################################################################################################################################################

#    def change_combo_ChildRamp(self,index_newvalue,index_combobox):
#        child_ramp_index = index_newvalue
#        index_row = index_combobox
#        if child_ramp_index ==  len(self.list_ramps): # This means that the value of the combobox is 'None'
#            self.experiment.disconnect_ramp_from_child(self.list_ramps[index_row][0])
#        else:
#            self.experiment.connect_ramp_to_child(self.list_ramps[index_row][0],self.list_ramps[child_ramp_index][0])
#        return


#################################
### END GUI Events Functions ####
#################################

############################################################################################################
## END status_selector                                                                                     # 
############################################################################################################


############################################################################################################
## ramp_connection                                                                                         # 
############################################################################################################

class ramp_connection():

    def __init__(self, parent, experiment):
        # parent        = a QWidget (or QMainWindow) object that will be the parent of this gui
        # experiment    = an experiment() object, whose acquisition will be controlled by this panel

        self.parent = parent
        self.experiment  = experiment 

        ### Connect signals from model to event slots of this GUI
        self.experiment.sig_list_ramps.connect(self.on_list_ramps_updated)
        #self.experiment.sig_single_set_acquisition.connect(self.on_single_set_acquisition_change)
        #self.experiment.sig_triggered_acquisition.connect(self.on_triggered_acquisition_change)
        #self.experiment.sig_trigger_modality.connect(self.on_trigger_modality_change)
        #self.experiment.sig_trigger_instrument.connect(self.on_trigger_instrument_change)
        #self.experiment.sig_time_interval_internal_trigger.connect(self.on_time_inteval_internal_trigger_change)
        #self.experiment.sig_time_interval_multiple_acquisitions_per_set.connect(self.on_time_interval_multiple_acquisitions_per_set_change)
        #self.experiment.sig_numb_acquisitions_per_set.connect(self.on_numb_acquisitions_per_trigger_change)
        #self.experiment.sig_make_average_set_acquisition.connect(self.on_change_make_average_set_acquisition)
        #self.experiment.sig_list_instruments.connect(self.on_list_instruments_updated)
        #self.experiment.sig_trigger_instrument.connect(self.on_trigger_instrument_changed)
        #self.experiment.sig_trigger_instrument_delay.connect(self.on_trigger_delay_changed)

        #self.doing_triggered_acquisition = False    # Flag used to keep track a triggered acquisition is being performed. If this is true, some widget will not be re-enabled when self.experiment.SIG_SINGLE_SET_ACQUISITION_ENDED is sent
        #                                            # but only when self.experiment.SIG_TRIGGERED_ACQUISITION_ENDED is sent

    def create_gui(self): 
        self.create_widgets()
        self.connect_widgets_events_to_functions()

    def create_widgets(self):
        """
        - hbox                                          (QHBoxLayout)

        """

        self.vbox = Qt.QVBoxLayout()
        self.label_Info = Qt.QLabel("This panal allows to connect a (parent) ramp to (child) ramp." \
                                     "The trigger event of the parent ramp becomes the start of the child ramp. " \
                                     "This can be used to sweep multiple parameters. Typically, the instrument of the child ramp " \
                                     "(or grandchild ramp, in case of multiple connections) would be set as 'instrument trigger' for the acquisition system.\n" \
                                     "\n<b>Important<b/>: the code does not do many sanity checks, and it will not complain about incestuous situations, where a ramp is set " \
                                     "as both a parent and a child of another ramp. This will generate unpredictable situations. Don't be incestuous.")
        self.label_Info.setWordWrap(True) 
        self.tableRamps = Qt.QTableWidget()
        self.tableRamps.setColumnCount(3)
        self.tableRamps.setHorizontalHeaderLabels(["Parent Instrument","Parent Ramp ID","Select child ramp:"])
        header = self.tableRamps.horizontalHeader()
        header.setSectionResizeMode(Qt.QHeaderView.ResizeToContents)       
        #header.setSectionResizeMode(0, Qt.QHeaderView.Stretch)
        self.vbox.addWidget(self.label_Info)
        self.vbox.addWidget(self.tableRamps)

        ### Assign the layout hbox to the wdiget defined in self.parent
        self.parent.setLayout(self.vbox) 
        self.parent.resize(self.parent.minimumSize())

        
        return
    
    def connect_widgets_events_to_functions(self):
        ### Connect Widget Events to Event Functions
        if hasattr(self,'list_combo_ramps'):
            for index,combo_box in enumerate(self.list_combo_ramps):
                self.list_combo_ramps[index].currentIndexChanged.connect(lambda index_newvalue, index=index : self.change_combo_ChildRamp(index_newvalue, index_combobox=index))
        return self

###########################################################################################################
### Event Slots. They are normally triggered by signals from the model, and change the GUI accordingly  ###
###########################################################################################################

    def on_list_ramps_updated(self,list_ramps):
        self.list_ramps = list_ramps #store for later use
        self.tableRamps.clearContents()

        # Create a list of ramps for the comboboxes 
        combobox_items = []
        for index,ramp in enumerate(list_ramps):
            combobox_items.append(f"{str(ramp[1])} -> {str(id(ramp[2]))}")
        combobox_items.append('None')

        self.list_combo_ramps = [] #this will contain the combobox for each row of the table
#
        self.tableRamps.setRowCount(len(list_ramps)) 
        for index,ramp in enumerate(list_ramps):
            item0 = Qt.QTableWidgetItem(str(ramp[1]))
            item1 = Qt.QTableWidgetItem(str(id(ramp[2])))
            self.tableRamps.setItem(index,0,item0)
            self.tableRamps.setItem(index,1,item1)
            combo_ramps = Qt.QComboBox()
            combo_ramps.addItems(combobox_items)
            self.tableRamps.setCellWidget(index,2,combo_ramps)
            self.list_combo_ramps.append(combo_ramps)
            
            #If the ramp corresponding to this row has child, we set the combobox to the corresponding child
            if ramp[3] == -1:
                combo_ramps.setCurrentIndex(len(combobox_items)-1) # This will initialize the combobox to 'None'
            else:
                combo_ramps.setCurrentIndex(ramp[3])

        #populating the table with combobox will create new widgets, thus we need to call again connect_widgets_events_to_functions
        self.connect_widgets_events_to_functions()


#######################
### END Event Slots ###
#######################

###################################################################################################################################################
### GUI Events Functions. They are triggered by direct interaction with the GUI, and they call methods of the experiment (i.e. the model) object.###
###################################################################################################################################################

    def change_combo_ChildRamp(self,index_newvalue,index_combobox):
        child_ramp_index = index_newvalue
        index_row = index_combobox
        if child_ramp_index ==  len(self.list_ramps): # This means that the value of the combobox is 'None'
            self.experiment.disconnect_ramp_from_child(self.list_ramps[index_row][0])
        else:
            self.experiment.connect_ramp_to_child(self.list_ramps[index_row][0],self.list_ramps[child_ramp_index][0])
        return


#################################
### END GUI Events Functions ####
#################################

############################################################################################################
## END ramp_connection                                                                                     # 
############################################################################################################

############################################################################################################
## acquisition_control                                                                                     # 
############################################################################################################

class acquisition_control():

    def __init__(self, parent, experiment):
        # parent        = a QWidget (or QMainWindow) object that will be the parent of this gui
        # experiment    = an experiment() object, whose acquisition will be controlled by this panel

        self.parent = parent
        self.experiment  = experiment 

        ### Connect signals from model to event slots of this GUI
        self.experiment.sig_single_set_acquisition.connect(self.on_single_set_acquisition_change)
        self.experiment.sig_triggered_acquisition.connect(self.on_triggered_acquisition_change)
        self.experiment.sig_trigger_modality.connect(self.on_trigger_modality_change)
        self.experiment.sig_trigger_instrument.connect(self.on_trigger_instrument_change)
        self.experiment.sig_time_interval_internal_trigger.connect(self.on_time_inteval_internal_trigger_change)
        self.experiment.sig_time_interval_multiple_acquisitions_per_set.connect(self.on_time_interval_multiple_acquisitions_per_set_change)
        self.experiment.sig_numb_acquisitions_per_set.connect(self.on_numb_acquisitions_per_trigger_change)
        self.experiment.sig_make_average_set_acquisition.connect(self.on_change_make_average_set_acquisition)
        self.experiment.sig_list_instruments.connect(self.on_list_instruments_updated)
        self.experiment.sig_trigger_instrument.connect(self.on_trigger_instrument_changed)
        self.experiment.sig_trigger_instrument_delay.connect(self.on_trigger_delay_changed)

        self.doing_triggered_acquisition = False    # Flag used to keep track a triggered acquisition is being performed. If this is true, some widget will not be re-enabled when self.experiment.SIG_SINGLE_SET_ACQUISITION_ENDED is sent
                                                    # but only when self.experiment.SIG_TRIGGERED_ACQUISITION_ENDED is sent

    def create_gui(self): 
        self.create_widgets()
        self.connect_widgets_events_to_functions()

    def create_widgets(self):
        """
        - hbox_acquisition_control                                          (QHBoxLayout)
            - box_trigger                                                   (QGroupBox)
                - box_trigger_hbox                                          (QVBoxLayout)
                    - StartPauseTriggeredAcquisition_hbox                   (QHBoxLayout)
                            ...widgets...
                    - InternalExternalTrigger_hbox                          (QHBoxLayout)
                        - panelInternalTrigger                              (QWidget)
                            - panelInternalTrigger_layout                   (QHBoxLayout)
                                ...widgets... 
                        - panelExternalTrigger                              (QWidget)
                            - panelExternalTrigger_layout                   (QVBoxLayout) 
                                - panelExternalTrigger_layout_row1          (QHBoxLayout)
                                    ...widgets... 
                                - panelExternalTrigger_layout_row2          (QHBoxLayout)
                                    ...widgets...
            - label_Arrow                                                   (QLabel)
            - box_acquisitionset                                            (QGroupBox)
                - box_acquisitionset_vbox                                   (QVBoxLayout)
                    - AcquisitionSet_row1                                   (QHBoxLayout)
                        ...widgets...
                    - AcquisitionSet_row2                                   (QHBoxLayout)
                        ...widgets...

        """

        hbox_acquisition_control = Qt.QHBoxLayout()
        box_trigger = Qt.QGroupBox()
        box_trigger.setTitle(f"Trigger")
        self.label_Arrow = Qt.QLabel() #This hosts the arrow image
        box_acquisitionset = Qt.QGroupBox()
        box_acquisitionset.setTitle(f"Acquisition(s)")

        #Initial setup of the three main blocks
        box_trigger.setObjectName('box_trigger')
        string = 'QGroupBox#box_trigger'
        box_trigger.setStyleSheet(string + ergastirio.styles.acquisitionpanel_box + '\nQGroupBox' + ergastirio.styles.acquisitionpanel_boxtitle ) #This line changes the style of ONLY this QWdiget
    
        pixmap = QtGui.QPixmap(os.path.join(graphics_dir,'arrow.png'))
        smaller_pixmap = pixmap.scaled(32, 32, QtCore.Qt.KeepAspectRatio, QtCore.Qt.FastTransformation)
        self.label_Arrow.setPixmap(smaller_pixmap)

        box_acquisitionset.setObjectName('box_acquisitionset')
        string = 'QGroupBox#box_acquisitionset'
        box_acquisitionset.setStyleSheet(string + ergastirio.styles.acquisitionpanel_box + '\nQGroupBox' + ergastirio.styles.acquisitionpanel_boxtitle ) #This line changes the style of ONLY this QWdiget

        hbox_acquisition_control.addWidget(box_trigger)
        hbox_acquisition_control.addWidget(self.label_Arrow)  
        hbox_acquisition_control.addWidget(box_acquisitionset)  
        #hbox_acquisition_control.addStretch(1)

        self.widgets_box_trigger = [] #This list will contain all widgets contained in box_trigger
        self.widgets_box_trigger_internal = []
        self.widgets_box_trigger_external = []
        self.widgets_box_acquisitionset = [] #This list will contain all widgets contained in box_acquisitionset
        self.widgets_box_acquisitionset_first_row = []

###### POPULATE box_trigger
        ### Add a horizontal box inside box_trigger
        box_trigger_hbox = Qt.QHBoxLayout()
        box_trigger.setLayout(box_trigger_hbox)

        ### Add two boxes, one horizontal to host the button and one vertical to host the two radio buttons
        StartPauseTriggeredAcquisition_hbox = Qt.QHBoxLayout()
        InternalExternalTrigger_hbox = Qt.QVBoxLayout()
        #vbox_trigger_box_vbox.setContentsMargins(0,0,0,0)
        box_trigger_hbox.addLayout(StartPauseTriggeredAcquisition_hbox)
        box_trigger_hbox.addLayout(InternalExternalTrigger_hbox)
        
        ### Populate the InternalExternalTrigger_hbox box with two widgets (in order to assign a style)
        self.panelInternalTrigger = Qt.QWidget()
        self.panelExternalTrigger = Qt.QWidget()
        InternalExternalTrigger_hbox.addWidget(self.panelInternalTrigger)
        InternalExternalTrigger_hbox.addWidget(self.panelExternalTrigger)
        
        ### Populate StartPauseTriggeredAcquisition_hbox with widgets
        StartPauseTriggeredAcquisition_hbox.setContentsMargins(0,0,0,0)
        StartPauseTriggeredAcquisition_hbox.setAlignment(QtCore.Qt.AlignCenter)
        self.button_StartPauseTriggeredAcquisition = Qt.QPushButton()
        self.button_StartPauseTriggeredAcquisition.setToolTip('Start or pause triggered data acquisition. Each time a trigger (either internal or by instrument) is received, \nit takes a single set of acquisitions.') 
        self.button_StartPauseTriggeredAcquisition.setText('Pause triggered\nacquisition')
        for w in [self.button_StartPauseTriggeredAcquisition]:
            StartPauseTriggeredAcquisition_hbox.addWidget(w)
            self.widgets_box_trigger.append(w)
        
        #panelStartPauseTriggeredAcquisition_layout.addStretch(1)

        ### Populate self.panelInternalTrigger with a hbox layout, and add widgets to the layout
        self.panelInternalTrigger.setStyleSheet("background-color: #D3D3D3;")
        panelInternalTrigger_layout = Qt.QHBoxLayout(self.panelInternalTrigger)
        panelInternalTrigger_layout.setContentsMargins(0,0,0,0)
        self.radio_TriggerInternal = Qt.QRadioButton()
        self.radio_TriggerInternal.setText("Internal")
        self.radio_TriggerInternal.setStyleSheet("QRadioButton { font: bold;}");
        self.radio_TriggerInternal.setChecked(True)
        self.radio_TriggerInternal.value= "internal"
        self.radio_TriggerInternal.setToolTip('In this modality, data from all instruments is acquired at periodic intervals of time, set by the \'refresh time\'. \nNote: user must check that each instrument is running and refreshing its own data.') 
        self.label_RefreshTime = Qt.QLabel("Time interval (s) = ")
        self.edit_RefreshTime = Qt.QLineEdit()
        self.edit_RefreshTime.setFixedWidth(50)
        
        self.edit_RefreshTime.setAlignment(QtCore.Qt.AlignRight)
        self.edit_RefreshTime.setStyleSheet("background-color: #FFFFFF;")
        for w in [self.radio_TriggerInternal,self.label_RefreshTime,self.edit_RefreshTime]:
            panelInternalTrigger_layout.addWidget(w)
            self.widgets_box_trigger.append(w)
            self.widgets_box_trigger_internal.append(w)
        panelInternalTrigger_layout.addStretch(1)

        ### Populate self.panelExternalTrigger with a vbox layout, which in turns contain two hbox layouts, and add widgets the hbox layouts
        self.panelExternalTrigger.setStyleSheet("background-color: #D3D3D3;")
        panelExternalTrigger_layout = Qt.QVBoxLayout(self.panelExternalTrigger)
        panelExternalTrigger_layout.setContentsMargins(0,0,0,0)
        panelExternalTrigger_layout_row1 = Qt.QHBoxLayout()
        panelExternalTrigger_layout_row2 = Qt.QHBoxLayout()
        panelExternalTrigger_layout.addLayout(panelExternalTrigger_layout_row1)
        panelExternalTrigger_layout.addLayout(panelExternalTrigger_layout_row2)
        self.radio_TriggerExternal = Qt.QRadioButton()
        self.radio_TriggerExternal.setText("By Instrument")
        self.radio_TriggerExternal.setStyleSheet("QRadioButton { font: bold;}");
        self.radio_TriggerExternal.value= "external"
        self.radio_TriggerExternal.setToolTip('In this modality, one instrument is designated as trigger. Whenever that instrument updates its own data, it triggers an acquisition.') 
        self.label_TriggerExternal = Qt.QLabel("=")
        self.combo_TriggerInstruments = Qt.QComboBox()
        self.combo_TriggerInstruments.resize(self.combo_TriggerInstruments.sizeHint())
        self.combo_TriggerInstruments.setStyleSheet("background-color: #FFFFFF;")
        self.label_TriggerInstrumentDelay = Qt.QLabel("[ delayed  by (s) = ")
        self.label_TriggerInstrumentDelay.setFont(QtGui.QFont('Arial', 8))
        self.edit_TriggerInstrumentDelay = Qt.QLineEdit()
        self.edit_TriggerInstrumentDelay.setStyleSheet("background-color: #FFFFFF;")
        self.edit_TriggerInstrumentDelay.setFixedWidth(50)
        self.edit_TriggerInstrumentDelay.setAlignment(QtCore.Qt.AlignRight)
        self.edit_TriggerInstrumentDelay.setFont(QtGui.QFont('Arial', 8))
        self.edit_TriggerInstrumentDelay.setFixedHeight(15)
        self.label_TriggerInstrumentDelay2 = Qt.QLabel("]")
        self.label_TriggerInstrumentDelay2.setFont(QtGui.QFont('Arial', 8))
        for w in [self.radio_TriggerExternal,self.label_TriggerExternal,self.combo_TriggerInstruments]:
            panelExternalTrigger_layout_row1.addWidget(w)
            self.widgets_box_trigger.append(w)
            self.widgets_box_trigger_external.append(w)
        panelExternalTrigger_layout_row1.addStretch(1)
        for w in [self.label_TriggerInstrumentDelay,self.edit_TriggerInstrumentDelay,self.label_TriggerInstrumentDelay2]:
            panelExternalTrigger_layout_row2.addWidget(w)
            self.widgets_box_trigger.append(w)
            self.widgets_box_trigger_external.append(w)
        panelExternalTrigger_layout_row2.addStretch(1)

        #Connect Trigger buttons
        self.buttongroup_trigger = Qt.QButtonGroup(self.parent)
        self.buttongroup_trigger.addButton(self.radio_TriggerInternal)
        self.buttongroup_trigger.addButton(self.radio_TriggerExternal)

###### POPULATE box_acquisitionset
        ### Add a vertical box inside box_acquisitionset
        box_acquisitionset_vbox = Qt.QVBoxLayout()
        box_acquisitionset_vbox.setAlignment(QtCore.Qt.AlignCenter)
        box_acquisitionset.setLayout(box_acquisitionset_vbox)

        ### Add two hbox layouts to box_acquisitionset_vbox
        AcquisitionSet_row1 =  Qt.QHBoxLayout()
        AcquisitionSet_row2 =  Qt.QHBoxLayout()
        box_acquisitionset_vbox.addLayout(AcquisitionSet_row1)
        box_acquisitionset_vbox.addLayout(AcquisitionSet_row2)

        ### Populate AcquisitionSet_row1 with widgets
        self.label_NumbAcquisitions = Qt.QLabel("Take N = ")
        self.spinbox_NumbAcquisitions = Qt.QSpinBox()
        self.spinbox_NumbAcquisitions.setRange(1, 10000)
        self.label_NumbAcquisitions2 = Qt.QLabel("acquisitions, every ")
        self.edit_AcquisitionRefreshTime = Qt.QLineEdit()
        self.edit_AcquisitionRefreshTime.setFixedWidth(30)
        self.edit_AcquisitionRefreshTime.setAlignment(QtCore.Qt.AlignRight)
        self.edit_AcquisitionRefreshTime.setStyleSheet("background-color: #FFFFFF;")
        self.label_NumbAcquisitions3 = Qt.QLabel("s,  ")
        self.box_MakeAverage = Qt.QCheckBox("and average them.")
        self.box_MakeAverage.setToolTip('When checked, the multiple acquisitions are averaged.')
        for w in [self.label_NumbAcquisitions, self.label_NumbAcquisitions, self.spinbox_NumbAcquisitions, self.label_NumbAcquisitions2, 
                  self.edit_AcquisitionRefreshTime, self.label_NumbAcquisitions3,self.box_MakeAverage]:
            AcquisitionSet_row1.addWidget(w)
            self.widgets_box_acquisitionset.append(w)
            self.widgets_box_acquisitionset_first_row .append(w)
        
        ### Populate AcquisitionSet_row2 with widgets
        self.button_SingleAcquisition = Qt.QPushButton("Take single set acquisitions")
        self.button_SingleAcquisition.setToolTip('Triggers a single set of acquisitions, with the number of acquisitions N and time interval specified here above.') 
        AcquisitionSet_row2.setAlignment(QtCore.Qt.AlignCenter)
        for w in [self.button_SingleAcquisition]:   
            AcquisitionSet_row2.addWidget(w)    
            self.widgets_box_acquisitionset.append(w)

        # Widgets for which we want to constraint the min size by using sizeHint()
        widget_list = [self.button_StartPauseTriggeredAcquisition,self.button_SingleAcquisition]
        for w in widget_list:
            w.setMinimumSize(w.sizeHint())

        # Widgets for which we want to constraint the max size by using sizeHint()
        #widget_list = [box_acquisitionset,box_trigger]
        #for w in widget_list:
        #    w.setMaximumSize(w.sizeHint())

        ### Assign the layout hbox_acquisition_control to the wdiget defined in self.parent
        self.parent.setLayout(hbox_acquisition_control) 
        self.parent.resize(self.parent.minimumSize())

        return
    
    def connect_widgets_events_to_functions(self):
        ### Connect Widget Events to Event Functions
        self.button_StartPauseTriggeredAcquisition.clicked.connect(self.click_button_StartPauseTriggeredAcquisition)
        self.radio_TriggerInternal.clicked.connect(self.click_radio_trigger)
        self.radio_TriggerExternal.clicked.connect(self.click_radio_trigger)
        self.edit_TriggerInstrumentDelay.returnPressed.connect(self.press_enter_edit_TriggerInstrumentDelay)
        self.edit_RefreshTime.returnPressed.connect(self.press_enter_edit_RefreshTime)
        self.combo_TriggerInstruments.currentIndexChanged.connect(self.change_combo_TriggerInstruments)
        self.button_SingleAcquisition.clicked.connect(self.click_button_SingleAcquisition)
        self.spinbox_NumbAcquisitions.valueChanged.connect(self.value_changed_spinbox_NumbAcquisitions)
        self.edit_AcquisitionRefreshTime.returnPressed.connect(self.press_enter_edit_AcquisitionRefreshTime)
        self.box_MakeAverage.stateChanged.connect(self.click_box_MakeAverage)
        return self

###########################################################################################################
### Event Slots. They are normally triggered by signals from the model, and change the GUI accordingly  ###
###########################################################################################################

    def on_single_set_acquisition_change(self,status):
        if self.doing_triggered_acquisition == False: #If we are doing a triggered acquisition, we dont change anything in the GUI when a single set of acquisitions is either started or ended
            if status == self.experiment.SIG_SINGLE_SET_ACQUISITION_STARTED:
                ergastirio.utils.disable_widget(self.widgets_box_trigger)
                ergastirio.utils.disable_widget(self.widgets_box_acquisitionset_first_row)
                self.button_SingleAcquisition.setText('Stop')
            if status == self.experiment.SIG_SINGLE_SET_ACQUISITION_ENDED:
                ergastirio.utils.enable_widget(self.widgets_box_trigger)
                ergastirio.utils.enable_widget(self.widgets_box_acquisitionset_first_row)
                self.button_SingleAcquisition.setText('Take single set acquisitions')

    def on_triggered_acquisition_change(self,status):
        if status == self.experiment.SIG_TRIGGERED_ACQUISITION_STARTED:
            self.doing_triggered_acquisition = True
            ergastirio.utils.enable_widget(self.widgets_box_trigger)
            ergastirio.utils.disable_widget(self.widgets_box_trigger_internal + self.widgets_box_trigger_external )
            ergastirio.utils.disable_widget(self.widgets_box_acquisitionset)
            self.button_StartPauseTriggeredAcquisition.setText('Pause triggered\nacquisition')
            pass
        if status == self.experiment.SIG_TRIGGERED_ACQUISITION_ENDED:
            self.doing_triggered_acquisition = False
            ergastirio.utils.enable_widget(self.widgets_box_trigger)
            ergastirio.utils.enable_widget(self.widgets_box_acquisitionset)
            self.button_StartPauseTriggeredAcquisition.setText('Start triggered\nacquisition')
            pass

    def on_trigger_modality_change(self,status):
        if status == self.experiment.SIG_TRIGGERED_MODALITY_INTERNAL:
            self.radio_TriggerInternal.setChecked(True)
            ergastirio.utils.enable_widget([self.edit_RefreshTime,self.label_RefreshTime])
            ergastirio.utils.disable_widget([self.combo_TriggerInstruments,self.edit_TriggerInstrumentDelay,self.label_TriggerExternal,self.label_TriggerInstrumentDelay])
        if status == self.experiment.SIG_TRIGGERED_MODALITY_EXTERNAL:
            self.radio_TriggerExternal.setChecked(True)
            ergastirio.utils.disable_widget([self.edit_RefreshTime,self.label_RefreshTime])
            ergastirio.utils.enable_widget([self.combo_TriggerInstruments,self.edit_TriggerInstrumentDelay,self.label_TriggerExternal,self.label_TriggerInstrumentDelay])

    def on_trigger_instrument_change(self,trigger_instrument):
        pass

    def on_time_inteval_internal_trigger_change(self,time_interval):
        self.edit_RefreshTime.setText(f"{time_interval}") 
        pass

    def on_time_interval_multiple_acquisitions_per_set_change(self,time_interval):
        self.edit_AcquisitionRefreshTime.setText(f"{time_interval}") 
        pass

    def on_numb_acquisitions_per_trigger_change(self,numb_acquisition):
        self.spinbox_NumbAcquisitions.setValue(numb_acquisition)
        pass

    def on_change_make_average_set_acquisition(self,status):
        self.box_MakeAverage.setChecked(status)

    def on_list_instruments_updated(self,list_instruments):
        self.combo_TriggerInstruments.blockSignals(True)
        self.combo_TriggerInstruments.clear()
        self.combo_TriggerInstruments.addItems(list_instruments) 
        self.combo_TriggerInstruments.blockSignals(False)

    def on_trigger_instrument_changed(self,instrument_index):
        self.combo_TriggerInstruments.blockSignals(True)
        self.combo_TriggerInstruments.setCurrentIndex(instrument_index)
        self.combo_TriggerInstruments.blockSignals(False)

    def on_trigger_delay_changed(self,delay):
        self.edit_TriggerInstrumentDelay.setText(f"{delay}") 

#######################
### END Event Slots ###
#######################

###################################################################################################################################################
### GUI Events Functions. They are triggered by direct interaction with the GUI, and they call methods of the experiment (i.e. the model) object.###
###################################################################################################################################################

    def click_button_StartPauseTriggeredAcquisition(self): 
        if(self.experiment.triggered_acquisition == False):
            self.press_enter_edit_AcquisitionRefreshTime()
            self.press_enter_edit_RefreshTime()
            #self.press_enter_edit_TriggerInstrumentDelay()
            self.experiment.start_triggered_acquisition()
        elif (self.experiment.triggered_acquisition == True):
            self.experiment.stop_triggered_acquisition()
        return

    def click_button_SingleAcquisition(self):
        if(self.experiment.single_set_acquisition == False):
            self.press_enter_edit_AcquisitionRefreshTime()
            self.experiment.start_single_set_acquisitions()
        elif (self.experiment.single_set_acquisition == True): #When this is true, the experiment is currently taking a single acquisition
            self.experiment.stop_single_set_acquisitions()
        return

    def press_enter_edit_RefreshTime(self):
        refresh_time_internal_trigger = self.edit_RefreshTime.text()
        self.experiment.time_interval_internal_trigger = refresh_time_internal_trigger #When doing this assignment, the setter of self.experiment.time_interval_internal_trigger will check if the input is valid, and eventually update the value
        return True
 
    def press_enter_edit_AcquisitionRefreshTime(self):
        acquisition_refresh_time = self.edit_AcquisitionRefreshTime.text()
        self.experiment.time_interval_multiple_acquisitions_per_set = acquisition_refresh_time #When doing this assignment, the setter of self.experiment.time_interval_multiple_acquisitions_per_set will check if the input is valid, and eventually update the value
        return True

    def press_enter_edit_TriggerInstrumentDelay(self):
        self.change_combo_TriggerInstruments()

    def value_changed_spinbox_NumbAcquisitions(self):
        self.experiment.numb_acquisitions_per_set = self.spinbox_NumbAcquisitions.value()

    def click_radio_trigger(self):
        if self.radio_TriggerInternal.isChecked():
            self.experiment.trigger_modality = 'internal'
            self.press_enter_edit_RefreshTime()
        if self.radio_TriggerExternal.isChecked():
            self.experiment.trigger_modality = 'external'
            self.change_combo_TriggerInstruments()
            
    def change_combo_TriggerInstruments(self):
        trigger_delay = self.edit_TriggerInstrumentDelay.text()
        self.experiment.set_trigger_instrument(self.combo_TriggerInstruments.currentIndex(),delay=trigger_delay)
        return

    def click_box_MakeAverage(self, state):
        if state == QtCore.Qt.Checked:
            status_bool = True
        else:
            status_bool = False
        self.experiment.make_average_set_acquisitions = status_bool

#################################
### END GUI Events Functions ####
#################################

############################################################################################################
## END acquisition_control                                                                                 # 
############################################################################################################




############################################################################################################
## data_management                                                                                         # 
############################################################################################################

class data_management():
    """
    Panel which contains the buttons 'Save Data','Delete all data','Delete last row'
    """
    def __init__(self, parent, experiment):
        """
        parent        = a QWidget (or QMainWindow) object that will be the parent of this gui
        experiment    = an experiment() object, whose acquisition will be controlled by this panel
        """
        self.parent = parent
        self.experiment  = experiment 

    def create_gui(self): 
        vbox1 = Qt.QVBoxLayout()

        self.button_SaveData = Qt.QPushButton("Save data")
        self.button_SaveData.setToolTip('Save all currently stored data in a .csv file.') 
        self.button_SaveData.clicked.connect(self.click_button_SaveData)
        self.button_DeleteAllData = Qt.QPushButton("Delete all data")
        self.button_DeleteAllData.setToolTip('Delete all currently stored data.') 
        self.button_DeleteAllData.clicked.connect(self.click_button_DeleteAllData)
        self.button_DeleteLastRowData = Qt.QPushButton("Delete last row")
        self.button_DeleteLastRowData.setToolTip('Delete last row of stored data.') 
        self.button_DeleteLastRowData.clicked.connect(self.click_button_DeleteLastRowData)

        vbox1.addWidget(self.button_SaveData)
        vbox1.addWidget(self.button_DeleteAllData)
        vbox1.addWidget(self.button_DeleteLastRowData)
        #vbox1.addStretch(1)

        vbox = Qt.QVBoxLayout()
        vbox.addLayout(vbox1)  
        #vbox.addStretch(1)

        self.parent.setLayout(vbox) #This line makes sure that all widgets defined so far are assigned to the widget defines in self.parent
        self.parent.resize(self.parent.minimumSize())

        return self

    ### GUI Events Functions

    def click_button_SaveData(self): 
        filename, _ = Qt.QFileDialog.getSaveFileName(self.parent, 
                        "Save File", "", "Csv Files (*.csv);;Text Files (*.txt)")
        if filename:
            self.experiment.save_stored_data(filename)
        return

    def click_button_DeleteAllData(self):
        answer = Qt.QMessageBox.question(self.parent,'', "Are you sure?", Qt.QMessageBox.Yes | Qt.QMessageBox.No, Qt.QMessageBox.No)
        if answer == Qt.QMessageBox.Yes:
            self.experiment.delete_current_data()
        return

    def click_button_DeleteLastRowData(self):
        self.experiment.delete_row_from_data(-1)
        return


instrument_box = """{
            font: bold;
            border: 1px solid black;
            border-radius: 6px;
            margin-top: 6px;
        }
"""

instrument_boxtitle = """::title {
            subcontrol-origin: margin;
            left: 7px;
            padding: -4px 5px 0px 5px;
        }
        """

acquisitionpanel_box = """{
            font: bold;
            border: 1px solid black;
            border-radius: 6px;
            margin-top: 6px;
        }
"""

acquisitionpanel_boxtitle = """::title {
            subcontrol-origin: margin;
            left: 7px;
            padding: -4px 5px 0px 5px;
        }
        """


import ergastirio.experiment_initializer
import PyQt5.QtCore as QtCore
import logging
import numpy as np
import datetime
import ergastirio.utils
import os
import json
import abstract_instrument_interface
## We follow as much as possible the model-view-controller paradigm. The class experiment() defines a model. The model emits signals (defined as QtCore.pyqtSignal object) 
# to let the different parts of the GUI know what is happening

class experiment(QtCore.QObject):
    """
    ......
    ...

    Attributes
    ----------
    Data
        instance of ergastirio.classes.EnhancedList
    verbose
        Defined as a @property. Can be set.
    
    name_logger
        Defined as a @property. Can be set.

    config_file
        Defined as a @property. Can be set. Changing its value will automatically load the new experiment settings.
    
    time_interval_internal_trigger
        Defined as a @property. Can be set.
    
    time_interval_multiple_acquisitions_per_set
        Defined as a @property. Can be set.
    
    numb_acquisitions_per_set
        Defined as a @property. Can be set.

    trigger_modality
        Defined as a @property. Can be set.

    make_average_set_acquisitions
        Defined as a @property. Can be set.

    settings = {
        "General_Settings": {   
                    "time_interval_internal_trigger": 0.2,              #in s. Refresh time of internal trigger (when used)
                    "numb_acquisitions_per_set": 1,                     #Every time that a trigger is "fired" (either by internal trigger, or by external trigger, or by clicking on 'take single acqusitions') 
                    "time_interval_multiple_acquisitions_per_set": 0.2, #the software takes a number of acquisitions, equal to numb_acquisitions_per_set, and separated in time by time_interval_multiple_acquisitions_per_set
                    "average_acquisition": 0,                           #When this is set to true, after taking a number of acquisitions (numb_acquisitions_per_set), it averages them and replace them by the average
                    "trigger_delay": 0,
                    "trigger_modality": 'internal',                     # can be either 'internal' or 'external'
                    "make_average_set_acquisitions": 0,                              #0 -> No Average, 1 -> Make Average
                    "trigger_instrument" : ''
                    }
                }

    data_headers

    single_set_acquisition 
    
    continuous_acquisition

    -------


    Methods 
    -------
    set_trigger_modality(modality)

    set_average_set_acquisition(status)

    read_current_data_from_all_instruments()

    store_current_data_from_all_instruments()

    [TO FINISH]

    """

    ## SIGNALS THAT WILL BE USED TO COMMUNICATE WITH THE GUI
    #                                                                           | Triggered when ...                                        | Sends as parameter    
    #                                                         #                 -----------------------------------------------------------------------------------------------------------------------         
    sig_single_set_acquisition = QtCore.pyqtSignal(int)                        #|                                                           |     
    sig_triggered_acquisition = QtCore.pyqtSignal(int)                         #|                                                           |  
    sig_trigger_modality = QtCore.pyqtSignal(str) 
    sig_trigger_instrument = QtCore.pyqtSignal(int)
    sig_trigger_instrument_delay = QtCore.pyqtSignal(float) 
    sig_list_instruments = QtCore.pyqtSignal(list) 
    sig_list_ramps = QtCore.pyqtSignal(list)                                   #|                                                           | List of all ramps defined in all instruments. The format is [ [id_instrument, fullname_instrument, ramp_object, index_ramp_child], [...] ... ] 
    sig_time_interval_internal_trigger = QtCore.pyqtSignal(float) 
    sig_time_interval_multiple_acquisitions_per_set = QtCore.pyqtSignal(float) 
    sig_numb_acquisitions_per_set = QtCore.pyqtSignal(int) 
    sig_make_average_set_acquisition = QtCore.pyqtSignal(int) 
    sig_data_updated = QtCore.pyqtSignal(int)                                  #| A generic data update has happened                        |
    sig_data_updated_added_new_rows = QtCore.pyqtSignal(int)                   #| A certain number of rows has been appended to the data    | Number of appended rows
    sig_data_updated_deleted_all_data = QtCore.pyqtSignal(int)                 #| All data has been deleted                                 |
    sig_data_updated_deleted_last_rows = QtCore.pyqtSignal(int)                #| A certain number of rows has been deleted from the data   | Number of deleted rows 
    
    
    
    ##
    # Identifier codes used for view-model communication. Other general-purpose codes are specified in abstract_instrument_interface
    SIG_SINGLE_SET_ACQUISITION_STARTED = 1
    SIG_SINGLE_SET_ACQUISITION_ENDED = 2
    SIG_TRIGGERED_ACQUISITION_STARTED = 1
    SIG_TRIGGERED_ACQUISITION_ENDED = 2
    SIG_TRIGGERED_MODALITY_INTERNAL = 'internal'
    SIG_TRIGGERED_MODALITY_EXTERNAL = 'external'
    SIG_AVERAGE_SET_ACQUISITION_ON = 1
    SIG_AVERAGE_SET_ACQUISITION_OFF = 2
    SIG_DATA_UPDATED_NEW_ROWS = 1
    SIG_DATA_UPDATED_DELETED_ALL_DATA = 2
    SIG_DATA_UPDATED_DELETED_LAST_ROWS = 3

    _verbose = True #Keep track of whether this instance of the interface should produce logs or not
    _name_logger = ''
    _config_file = ''

    def __init__(self, app, name_logger=__package__):
        # app           = The pyqt5 QApplication() object
        # name_logger   = The name of the logger used for this particular experiment. If none is specified, the name of the package (i.e. ergastirio) is used as logger name

        QtCore.QObject.__init__(self)
        self.app = app

        self.name_logger = name_logger #Setting this property will also create the logger,set output style, connect the logger output to the logging textarea and store the logger object in self.logger (see @name_logger.setter) 
        
        
        self.load_default_settings() ### Load default values of settings (might be overwritten by settings passed by user later)

        self.data_headers =[]   #List of strings, will contain the label of each 'column" of acquired data, based on the instruments connected, in the format "dev#i_dataname#j"
                                #where #i runs from 0 to the number of instruments minus 1, and dataname#j is the name of the jth data created by the i-th instrument.
                                #the data created by each instrument are specified by the keys of the 'output' dictionary defined in the interface of each instrument

        self.single_set_acquisition = False     #Boolean variable, true while performing a single-set acquisition
        self.triggered_acquisition = False      #Boolean variable, true when performing continuous acquisition (i.e. when internal trigger is used)

        return

    def load_default_settings(self):
        folder_default_file = os.path.dirname(os.path.abspath(__file__))
        config_file = os.path.join(folder_default_file,'default_settings.json')
        self.logger.info(f"Loading default settings...")
        try:
            with open(config_file) as jsonfile:
                self.settings = json.load(jsonfile)
        except Exception as e:
            self.logger.error(f"An error occurred while loading default settings from  {config_file}. Fix the error and restart this application\n: {e}")

    def initiate(self, config_file, func_logger_handler_adder = None, **kwargs):
        if ('virtual' in kwargs.keys()) and (kwargs['virtual'] == True):
            self.logger.info(f"Initializing the experiment with virtual instruments (when available)...")
            self.virtual_modality = True 
        else:    
            self.virtual_modality = False
        self.func_logger_handler_adder = func_logger_handler_adder
        self.func_logger_handler_adder(logger=logging.getLogger(self.name_logger)) #Having this here is kind of ugly because it partially breaks the model-view philosophy; on the other hand it allows catching the logs immediately
        self.config_file = config_file 
        self.load_settings()
        self.logger.info(f"Initializing the experiment...")
        return ergastirio.experiment_initializer.set_up_experiment(self)

    def load_settings(self):
        self.logger.info(f"Loading the content of {self.config_file}")
        try:
            with open(self.config_file) as jsonfile:
                settings = json.load(jsonfile)
                for key in settings.keys():
                    if key in self.settings.keys():
                        self.settings[key].update(settings[key])
                    else:
                        self.settings[key] = settings[key]
        except Exception as e:
            self.logger.error(f"An error occurred while loading the file {self.config_file}. Fix the error and restart this application\n: {e}")

    def save_settings(self):
        if self.config_file:
            self.logger.info(f"Storing current settings for this experiment into the file \'{self.config_file}\'...")
            #Collect settings dictionaries from all instruments
            list_settings_instruments = [instrument['interface'].settings for instrument in self.instruments]
            self.settings["Instruments_Settings"]["settings_each_instrument"] = list_settings_instruments
            try:
                with open(self.config_file, 'w') as fp:
                    json.dump(self.settings, fp, indent=4, sort_keys=True)
            except Exception as e:
                self.logger.error(f"An error occurred while saving settings in the file {self.config_file}: {e}")
        return

    def fire_all_signals(self):
        """
        Fire all signals, communicating the current value of several parameters. This function is typically called from the GUI of the experiment after the GUI has been partially
        initiallized. By making the experiment object fire all its parameters, the GUI can properly populate all of its field
        """
        self.sig_time_interval_internal_trigger.emit(self.settings["General_Settings"]['time_interval_internal_trigger'])
        self.sig_time_interval_multiple_acquisitions_per_set.emit(self.settings["General_Settings"]['time_interval_multiple_acquisitions_per_set'])
        self.sig_numb_acquisitions_per_set.emit(self.settings["General_Settings"]['numb_acquisitions_per_set'])
        self.sig_make_average_set_acquisition.emit(self.settings["General_Settings"]['make_average_set_acquisitions'])
        self.sig_triggered_acquisition.emit(self.SIG_TRIGGERED_ACQUISITION_ENDED)
        self.sig_single_set_acquisition.emit(self.SIG_SINGLE_SET_ACQUISITION_ENDED)
        self.sig_trigger_modality.emit(self.settings["General_Settings"]['trigger_modality'])
        self.sig_list_instruments.emit([instrument['fullname'] for instrument in self.instruments])
        self.sig_list_ramps.emit(self.create_list_ramps())
        self.sig_trigger_instrument.emit(self.settings["Instruments_Settings"]['trigger_instrument'])
        self.sig_trigger_instrument_delay.emit(self.settings["Instruments_Settings"]['trigger_delay'])

    def create_list_ramps(self):
        list_ramps = []
        for index, instrument in enumerate(self.instruments):
            if hasattr(instrument['interface'],'ramp') and isinstance(instrument['interface'].ramp, abstract_instrument_interface.ramp):
                list_ramps.append([index,instrument['fullname'],instrument['interface'].ramp])

        #Does another for loop to check if the ramps have children; if so, stores the index of the child ramp as 4th element of each list (-1 otherwise)
        for index, ramp in enumerate(list_ramps):
            list_ramps[index].append(-1)
            ramp_object = ramp[2]
            if hasattr(ramp_object,'child_ramp') and isinstance(ramp_object.child_ramp, abstract_instrument_interface.ramp):
                for index_possible_child,ramp_possible_child in enumerate(list_ramps):
                    if ramp_possible_child[2] == ramp_object.child_ramp:
                        list_ramps[index][3] = index_possible_child
                        break
        return list_ramps

    @property
    def verbose(self):
        return self._verbose
    @verbose.setter
    def verbose(self,verbose): #When the verbose property is changed, we also update accordingly the level of the logger object
        self._verbose = verbose
        if verbose: loglevel = logging.INFO
        else: loglevel = logging.CRITICAL
        self.logger.setLevel(level=loglevel)

    @property
    def name_logger(self):
        return self._name_logger
    @name_logger.setter
    def name_logger(self,name): #Create logger, and set default output style.
        self._name_logger = name
        self.logger = logging.getLogger(self._name_logger)
        self.verbose = self._verbose #This will automatically set the logger verbosity too
        if not self.logger.handlers:
            formatter = logging.Formatter(f"[{self.name_logger}]: %(message)s")
            ch = logging.StreamHandler()
            ch.setFormatter(formatter)
            self.logger.addHandler(ch)
        self.logger.propagate = False

    @property
    def config_file(self):
        return self._config_file
    @config_file.setter
    def config_file(self,config_file): #Setup a new config file, and then call the function ergastirio.experiment_initializer.load_config_and_setup_exp to initialize the experiment
        self.logger.info(f"Setting the config file to {config_file}...")
        self._config_file = config_file

    @property
    def time_interval_internal_trigger(self):
        return self.settings["General_Settings"]['time_interval_internal_trigger']
    @time_interval_internal_trigger.setter
    def time_interval_internal_trigger(self,r):
        try: 
            r = float(r)
            if r < 0.001:
                self.logger.error(f"The time interval of the internal trigger must be positive and >= 1ms.")
            else:
                if not self.settings["General_Settings"]['time_interval_internal_trigger'] == r:
                    self.settings["General_Settings"]['time_interval_internal_trigger'] = r
                    self.logger.info(f"The time interval of the internal trigger is now {r} s.")
        except ValueError:
            self.logger.error(f"The time interval of the internal trigger must be a valid number.")
        self.sig_time_interval_internal_trigger.emit(self.settings["General_Settings"]['time_interval_internal_trigger'])
        return self.settings["General_Settings"]['time_interval_internal_trigger']

    @property 
    def time_interval_multiple_acquisitions_per_set(self):
        return self.settings["General_Settings"]['time_interval_multiple_acquisitions_per_set']
    @time_interval_multiple_acquisitions_per_set.setter
    def time_interval_multiple_acquisitions_per_set(self,r):
        try: 
            r = float(r)
            if r < 0.001:
                self.logger.error(f"The time interval between multiple acquisitons must be positive and >= 1ms.")
            else:
                if not self.settings["General_Settings"]['time_interval_multiple_acquisitions_per_set'] == r:
                    self.settings["General_Settings"]['time_interval_multiple_acquisitions_per_set'] = r
                    self.logger.info(f"The time interval between multiple acquisitons is now {r} s.")
        except ValueError:
            self.logger.error(f"The time interval between multiple acquisitons must be a valid number.")
        self.sig_time_interval_multiple_acquisitions_per_set.emit(self.settings["General_Settings"]['time_interval_multiple_acquisitions_per_set'])
        return self.settings["General_Settings"]['time_interval_multiple_acquisitions_per_set']

    @property
    def numb_acquisitions_per_set(self):
        return self.settings["General_Settings"]['numb_acquisitions_per_set']
    @numb_acquisitions_per_set.setter
    def numb_acquisitions_per_set(self,n):
        try: 
            n = int(n)
            if n < 1:
                self.logger.error(f"The number of acquisitions per set must be a positive integer.")
            else:
                if not self.settings["General_Settings"]['numb_acquisitions_per_set'] == n:
                    self.settings["General_Settings"]['numb_acquisitions_per_set'] = n
                    self.logger.info(f"The number of acquisitions per set is now {n}.")
        except ValueError:
            self.logger.error(f"The number of acquisitions per set must be a positive integer.")
        self.sig_numb_acquisitions_per_set.emit(self.settings["General_Settings"]['numb_acquisitions_per_set'])
        return 

    @property
    def trigger_modality(self):
        return self.settings["General_Settings"]['trigger_modality']
    @trigger_modality.setter
    def trigger_modality(self,modality):
        if not (modality in ['internal','external']):
            self.logger.error(f"Trigger modality must be either 'internal' or 'external'.")
            self.reset_all_triggers() #Make sure to remove any trigger from all instruments
            return
        self.settings["General_Settings"]['trigger_modality'] = modality
        self.sig_trigger_modality.emit(modality)
        self.logger.info(f"Trigger modality set to {modality}.")

    @property
    def make_average_set_acquisitions(self):
        return bool(self.settings["General_Settings"]['make_average_set_acquisitions'])
    @make_average_set_acquisitions.setter
    def make_average_set_acquisitions(self,modality):
        if not (modality in [True, False]):
            self.logger.error(f"make_average_set_acquisitions must be either True or False.")
            return
        self.settings["General_Settings"]['make_average_set_acquisitions'] = modality
        self.sig_make_average_set_acquisition.emit(modality)
        self.logger.info(f"make_average_set_acquisitions set to {modality}.")

    def read_current_data_from_all_instruments(self,timestamp):
        '''
        Look into all the instruments interfaces (defined via the key 'interface' in each dictionary of the list exp.instruments) 
        and it extracts data from each instrument. From each instrument, the data to exctract is contained in the dictionary exp.instruments[i]['interface'].output
        '''
        current_data = []
        self.logger.info(f"Reading data from all instruments...")
        for instrument in self.instruments:
            instruments_data = instrument['interface'].read_current_output()
            instrument['interface'].receive_trigger(timestamp=timestamp)
            for data in instruments_data.values():
                current_data.append(data)  
        return current_data

    def store_current_data_from_all_instruments(self):
        '''
        '''
        now = datetime.datetime.now()
        time_string=now.strftime("%Y-%m-%d %H:%M:%S.%f")
        timestamp = datetime.datetime.timestamp(now)

        current_data = self.read_current_data_from_all_instruments(timestamp=timestamp) #Read the current data from all instruments
        acq_numb = len(self.data) + 1 #Calculate the number of rows of self.data, add 1 for the new row
        self.logger.info(f"[{time_string}] Acquisition #{acq_numb}")# {current_data}")
        current_data.insert(0, time_string)
        current_data.insert(0, timestamp )
        
        if ('post_processed_data' in self.settings['Instruments_Settings'].keys()):
            variables = {self.data_headers[i]: current_data[i] for i in range(len(current_data))}  #Create a dictionary with the values of the data that we just acquired
            for post_processed_data_name,post_processed_data_formula in self.settings['Instruments_Settings']['post_processed_data'].items():
                PostProcessedData = ergastirio.utils.evaluate_formula(post_processed_data_formula,variables)
                current_data.append(PostProcessedData)
        self.data.append(current_data)
        self.data_std.append([0]*(len(current_data)-2)) #Add a row of zeros to the standard deviations. The -2 is to avoid the columns corresponding to time and timestamp
        self.sig_data_updated_added_new_rows.emit(1)

    def start_single_set_acquisitions(self):
        self.single_set_acquisition = True
        self.sig_single_set_acquisition.emit(self.SIG_SINGLE_SET_ACQUISITION_STARTED)
        self.logger.info(f"Starting acquisition set...")
        self._single_set_of_acquisitions(counter=self.settings["General_Settings"]['numb_acquisitions_per_set'], time_interval=self.settings["General_Settings"]['time_interval_multiple_acquisitions_per_set'])

    def stop_single_set_acquisitions(self):
        self.single_set_acquisition = False
        self.sig_single_set_acquisition.emit(self.SIG_SINGLE_SET_ACQUISITION_ENDED)
        self.logger.info(f"Acquisition set stopped by user.")

    def _single_set_of_acquisitions(self,counter,time_interval):
        ''' Takes a single set of acquisitions. The number of acquisitions is  specified by the input counter, and they are separated by 
            a time interval specified by time_interval (in seconds).

            This function calls the function take_single_acquisition() and then calls itself after a time specified by self.settings['General_Settings']['time_interval_multiple_acquisitions_per_set'], 
            and with counter = counter -1. It also checks that self.single_set_acquisition = True, which can be used to interrupt the single set of measurements
            Normally, this function is called by another function with the input argument counter = self.settings['General_Settings']['numb_acquisitions_per_set']
        '''
        self.take_single_set_acquisition()
        if (counter > 1 and self.single_set_acquisition==True): 
            #if there are more acquisitions to be taken (counter>1) and if the measurment was not interrupted (self.single_set_acquisition==True), call the function again
            QtCore.QTimer.singleShot(int(time_interval*1e3), lambda: self._single_set_of_acquisitions(counter=counter-1,time_interval=time_interval))
        else:
            self.sig_single_set_acquisition.emit(self.SIG_SINGLE_SET_ACQUISITION_ENDED)
            self.single_set_acquisition=False
            self.logger.info(f"Acquisition set ended.")
            if self.settings["General_Settings"]['make_average_set_acquisitions'] and self.settings["General_Settings"]['numb_acquisitions_per_set']>1:
                self.logger.info(f"Averaging the last {self.settings['General_Settings']['numb_acquisitions_per_set']} acquisitions...")
                self.make_average_last_acquisitions(N=self.settings['General_Settings']['numb_acquisitions_per_set'])

    def take_single_set_acquisition(self):
        self.store_current_data_from_all_instruments()

    def make_average_last_acquisitions(self,N):
        '''
        Take the last N elements of self.data (where typically N = self.settings['General_Settings']['numb_acquisitions_per_set']) and replace them with their average (column-wise)
        It also populates the standard deviations (TO DO)
        '''
        rows_to_average = self.data[-N:]    #Get the last N rows
        timestamp = rows_to_average[0][0]   #Extract timestap and time of the first row to be averaged
        time = rows_to_average[0][1]
        rows_to_average2 = [elem[2:] for elem in rows_to_average] #Remove first two columns, which is timestamp and time
### TO DO: check if the list rows_to_average2 is indeed a 1D array of floats or integer
        averaged = (np.nanmean(np.array(rows_to_average2),axis=0)).tolist()
        std = (np.nanstd(np.array(rows_to_average2),axis=0)).tolist()
        averaged.insert(0, time)
        averaged.insert(0, timestamp )
        del self.data[-N:]      #Remove last N rows from data
        del self.data_std[-N:]  #Remove last N rows from data_std
        self.data.append(averaged)
        self.data_std.append(std)
        self.sig_data_updated.emit(0)

    def on_trigger_received(self):
        '''
        This function is invokeded either by the internal trigger, or by the instrument that act as a trigger. 
        Every time it is invoked, it takes a single set of acquisitions.
        '''
        self.start_single_set_acquisitions()

    def start_triggered_acquisition(self):
        if self.settings["General_Settings"]['trigger_modality'] == 'external':
            self.triggered_acquisition = True
            instrument_name = self.instruments[self.settings['Instruments_Settings']['trigger_instrument']]['fullname']
            self.logger.info(f"Triggered acquisition (external trigger) started...")
            self.logger.info(f"Waiting for trigger from {instrument_name}...")
            self.sig_triggered_acquisition.emit(self.SIG_TRIGGERED_ACQUISITION_STARTED)
        else:
            time_required_single_acquisition_set = (self.settings["General_Settings"]['numb_acquisitions_per_set']-1) * self.settings["General_Settings"]['time_interval_multiple_acquisitions_per_set']
            if self.settings["General_Settings"]['time_interval_internal_trigger'] <= time_required_single_acquisition_set:
                self.logger.error(f"Time interval of internal trigger ({self.settings['General_Settings']['time_interval_internal_trigger']} s) must be longer than the time required for a single acquisition set ({time_required_single_acquisition_set} s). Otherwise different acquisition sets would overlap in time, leading to unpredictable behaviour, and possibly the thermal death of the universe.")
                return
            self.triggered_acquisition = True
            self.sig_triggered_acquisition.emit(self.SIG_TRIGGERED_ACQUISITION_STARTED)
            self.logger.info(f"Triggered acquisition (internal trigger) started...")
            self.internal_trigger()
        return

    def stop_triggered_acquisition(self):
        self.triggered_acquisition = False
        self.sig_triggered_acquisition.emit(self.SIG_TRIGGERED_ACQUISITION_ENDED)
        self.logger.info(f"Triggered acquisition stopped.")

    def internal_trigger(self):
        '''
        Once called, this routine is executed continuosly, at time intervals set by self.settings["General_Settings"]['time_interval_internal_trigger']
        as long as (self.triggered_acquisition == True) and (self.settings["General_Settings"]['trigger_modality'] == 'internal').
        If the experiment is set in internal trigger modality and if self.triggered_acquisition == True, 
        the function self.update is called. Otherwise, none is done
        '''
        if (self.triggered_acquisition == True) and (self.settings["General_Settings"]['trigger_modality'] == 'internal'):
            self.on_trigger_received()
            QtCore.QTimer.singleShot(int(self.settings["General_Settings"]['time_interval_internal_trigger']*1e3), self.internal_trigger)  
                
    def on_external_trigger_received(self):
        '''
        This function is called by instruments if they are set as trigger and when they fire a trigger
        '''
        instrument_name = self.instruments[self.settings['Instruments_Settings']['trigger_instrument']]['fullname']
        if self.triggered_acquisition:
            if self.settings["General_Settings"]['trigger_modality'] == 'external':
                self.logger.info(f"Received trigger from an instrument...")
                self.on_trigger_received()
                #self.logger.info(f"Waiting for next trigger from {instrument_name}...")
            else:
                self.logger.info(f"Received a trigger from {instrument_name}, but experiment is not in 'external trigger' modality.")
        else:
            self.logger.info(f"Received a trigger from {instrument_name}, but experiment is not currently acquiring data.")

    def reset_all_triggers(self):
        for instrument in self.instruments: # First we remove the trigger from all instruments   
            try:
                instrument['interface'].set_trigger(None,delay=0)
            except Exception as e:
                self.logger.error(f"{e}")
        return

    def find_id_and_name_of_instrument(self,instrument):
        '''
        the input argument instrument can be either a number (indicating the position of the instrument in the list of instruments) or the full name of the instrument
        '''
        if isinstance(instrument,str):
            list_instruments = [inst['fullname'] for inst in self.instruments]
            if not (instrument in list_instruments):
                self.logger.error(f"{instrument} is not a valid instrument name")
                return False
            else:
                instrument_index = list_instruments.index(inst)
                instrument_name = instrument
        elif isinstance(instrument,int):
            if not(instrument in range(len(self.instruments))):
                self.logger.error(f"{instrument} is not a valid instrument index. Possible values are between 0 and {len(self.instruments) -1}.")
                return False
            else:
                instrument_index = instrument
                instrument_name = self.instruments[instrument]['fullname']
        else:
            self.logger.error(f"The input argument {instrument} must be either a string or an integer.")
            return False
        return instrument_index,instrument_name

    def set_trigger_instrument(self,instrument,delay=0):
        '''
        instrument can be either a number (indicating the position of the instrument in the list of instruments) or the full name of the instrument
        '''
        result = self.find_id_and_name_of_instrument(instrument)
        if not(result):
            return
        instrument_index,instrument_name = result

        try: 
            delay = float(delay)
        except ValueError:
            self.logger.error(f"The trigger delay must be a valid number.")
            return 
        if delay < 0:
            self.logger.error(f"The delay time must be positive and >= 1ms.")
            return 

        self.logger.info(f"Setting the instrument {self.instruments[instrument_index]['fullname']} as trigger, with delay = {delay} s...")
        self.reset_all_triggers()
        try:
            self.instruments[instrument_index]['interface'].set_trigger(self.on_external_trigger_received,delay=delay)
            self.settings['Instruments_Settings']['trigger_instrument'] = instrument_index
            self.settings['Instruments_Settings']['trigger_delay'] = delay
        except Exception as e:
            self.logger.error(f"An error occurred while setting {self.instruments[instrument_index]['fullname']} as trigger:\n {e}")
            return False

        self.logger.info(f"{self.instruments[instrument_index]['fullname']} was succesfully set as trigger.")
        self.sig_trigger_instrument.emit(self.settings["Instruments_Settings"]['trigger_instrument'])
        self.sig_trigger_instrument_delay.emit(self.settings['Instruments_Settings']['trigger_delay'])
        return True

    def connect_ramp_to_child(self,instrument_parent,instrument_child):
        result = self.find_id_and_name_of_instrument(instrument_parent)
        if not(result):
            return
        instrument1_index,instrument1_name = result
        result = self.find_id_and_name_of_instrument(instrument_child)
        if not(result):
            return
        instrument2_index,instrument2_name = result
        if instrument1_index == instrument2_index:
            self.logger.error(f"Cannot connect a ramp to itself.")
        else:
            self.instruments[instrument1_index]['interface'].ramp.connect_to_ramp_child(self.instruments[instrument2_index]['interface'].ramp)
            self.logger.info(f"The ramp of instrument {self.instruments[instrument1_index]['fullname']} was succesfully connected to the ramp of {self.instruments[instrument2_index]['fullname']}.")
        self.sig_list_ramps.emit(self.create_list_ramps())

    def disconnect_ramp_from_child(self,instrument_parent):
        result = self.find_id_and_name_of_instrument(instrument_parent)
        if not(result):
            return
        instrument1_index,instrument1_name = result
        self.instruments[instrument1_index]['interface'].ramp.disconnect_from_ramp_child()
        self.logger.info(f"The ramp of instrument {self.instruments[instrument1_index]['fullname']} was succesfully disconnected from its child.")
        self.sig_list_ramps.emit(self.create_list_ramps())

    def delete_current_data(self):
        self.logger.info(f"All store data was deleted.")
        self.data.clear()
        self.data_std.clear()
        self.sig_data_updated_deleted_all_data.emit(0)

    def delete_row_from_data(self,row):
        try:
            self.data.pop(row)
            self.data_std.pop(row)
            if row == -1:
                row = 'Last'
            self.logger.info(f"{row} row has been removed from data.")
            self.sig_data_updated_deleted_last_rows.emit(1)
        except:
            pass

    def save_stored_data(self,filename):
        '''
        Saves the values of all currently stored data on file. 
        The data are saved in a tabular form and delimited by a comma.
        '''
        d =','
        header = ''
        for h in self.data_headers:
            header = header + h + d 
        for h in self.data_headers: #We skip the first element of headers which corresponds to acquisition time
            header = header + (h+'_std') +  d
        if len(self.data)>0:
            A = np.concatenate((np.array(self.data), np.array(self.data_std)),axis=1)
            Nrows = A.shape[0]
        else:
            A = []
            Nrows = 0
        np.savetxt(filename, A, delimiter=d, header=header, comments="",fmt='%s')#,fmt=d.join(['%s'] + ['%e']*(A.shape[1]-1)))
        self.logger.info(f"Saved all stored data (number of rows = {Nrows}) in the file {filename}")
        return

    def close(self):
        self.save_settings()
        for instrument in self.instruments:
            try:
                instrument['interface'].close()
            except:
                pass




'''
This module contains several customized widgets used by Ergastirio
'''
import PyQt5.QtWidgets as Qt
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import numpy as np
import os

graphics_dir = os.path.join(os.path.dirname(__file__), '../graphics')

class Table():
    ''' Interactive table. 
    '''
    def __init__(self,  parent):
        self.parent = parent
        self._data = []
        self._data_headers = []
        self.table_widget = Qt.QTableWidget()
        #Qt.QTableWidget.__init__(self, *args)
        #self.verticalHeader().setVisible(False)
        self.refresh_continuosly = True

    def create_gui(self): 
        self.create_widgets()
        self.connect_widgets_events_to_functions()

    def create_widgets(self):
        self.vbox = Qt.QVBoxLayout()
        self.box_UpdateContinuosly = Qt.QCheckBox("Update table continuosly (might slow down measurements for large data).")
        self.box_UpdateContinuosly.setChecked(True)
        self.vbox.addWidget(self.box_UpdateContinuosly)
        self.vbox.addWidget(self.table_widget)

        ### Assign the layout hbox to the wdiget defined in self.parent
        self.parent.setLayout(self.vbox) 
        self.parent.resize(self.parent.minimumSize())

    def connect_widgets_events_to_functions(self):
        self.box_UpdateContinuosly.stateChanged.connect(self.click_box_UpdateContinuosly)

    @property
    def data_headers(self):
        return self._data_headers
    @data_headers.setter
    def data_headers(self,h):
        self._data_headers = h
        horHeaders = self._data_headers 
        self.table_widget.setColumnCount(len(self._data_headers))
        self.table_widget.setHorizontalHeaderLabels(horHeaders)
        self.table_widget.resizeColumnsToContents()

    @property
    def data(self):
        return self._data
    @data.setter
    def data(self,d):
        self._data = d
        rows = len(self._data)
        #self.setRowCount(rows)
        #for m,row in enumerate(self._data):
        #    for n,item in enumerate(row):
        #        newitem = Qt.QTableWidgetItem(str(row[n]))
        #        self.setItem(m, n, newitem)  
        #self.data_headers = self._data_headers #We need to call this after data is added
        #self.resizeRowsToContents()
        #self.scrollToBottom()

    def display_all_data(self):
        # Refreshs the whole table by looking at the current value of data
        self.on_rows_added(Nrows = len(self._data))

    def add_signals(self,
                    signal_data_updated,
                    signal_added_new_rows,
                    signal_deleted_all_data,
                    signal_deleted_last_rows):
        signal_data_updated.connect(self.display_all_data)
        signal_added_new_rows.connect(self.on_rows_added)
        signal_deleted_all_data.connect(self.on_delete_all_data) 
        signal_deleted_last_rows.connect(self.on_last_rows_deleted)
        return

    def on_delete_all_data(self,Nrows):
        #if self._data has been emptied and if on_rows_added() is called with input = 0, it will clean up the table
        temp = self.refresh_continuosly
        self.refresh_continuosly = True
        self.on_rows_added(Nrows)
        self.refresh_continuosly = temp

    def on_last_rows_deleted(self,Nrows):
        #if the last rows of self._data have been removed and on_rows_added() is called with a negative parameters, it will effectively remove the deleted rows from the table
        temp = self.refresh_continuosly
        self.refresh_continuosly = True
        self.display_all_data()
        self.on_rows_added(-Nrows)
        self.refresh_continuosly = temp

    def on_rows_added(self,Nrows):
        if self.refresh_continuosly == False:
            return
        total_rows = len(self._data)
        self.table_widget.setRowCount(total_rows)
        if Nrows > 0:
            rows_shown = total_rows - Nrows
            for m,row in enumerate(self._data[-Nrows:]):
                for n,item in enumerate(row):
                    newitem = Qt.QTableWidgetItem(str(row[n]))
                    self.table_widget.setItem(m+rows_shown, n, newitem)  
            self.data_headers = self._data_headers #We need to call this after data is added
            self.table_widget.resizeRowsToContents()
            self.table_widget.scrollToBottom()

###################################################################################################################################################
### GUI Events Functions.###
###################################################################################################################################################

    def click_box_UpdateContinuosly(self, state):
        if state == QtCore.Qt.Checked:
            self.refresh_continuosly = True
            self.display_all_data()
        else:
            self.refresh_continuosly = False

#################################
### END GUI Events Functions ####
#################################